import logging

from yt_dlpp.workers.worker import Worker


class DedupWorker(Worker[str, str]):
    """Worker in charge of deduplicating inputs"""

    _seen: set[str]

    def __init__(self, input_queue, output_queue) -> None:
        super().__init__(input_queue, output_queue)
        self._seen = set()

    def _process_item(self, item):
        if item in self._seen:
            logging.debug("Skipping duplicate item: %s", item)
            return
        logging.debug("Relaying item: %s", item)
        self._seen.add(item)
        self._send_output(item)



import json
import logging
from multiprocessing import JoinableQueue
from subprocess import CalledProcessError, run
from typing import Sequence

from yt_dlpp.interceptors.interceptor import InfoInterceptor
from yt_dlpp.workers.worker import Worker


class InfoWorker(Worker[str, str]):
    """
    Worker process that treats yt-dlp urls, gets info from them and passes video urls.

    - The input url may refer to a video or playlist.
    """

    input_queue: JoinableQueue
    output_queue: JoinableQueue

    _base_command: Sequence[str]

    def __init__(
        self,
        ydl_args: Sequence[str],
        input_queue: JoinableQueue,
        output_queue: JoinableQueue,
    ) -> None:
        super().__init__(input_queue, output_queue)
        # Define base command args
        interceptor = InfoInterceptor()
        _, allowed = interceptor.parse_known_args(ydl_args)
        self._base_command = (
            "yt-dlp",
            "--simulate",
            "--dump-json",
            *allowed,
        )
        logging.debug(
            "InfoWorker initialised with base command: %s",
            " ".join(self._base_command),
        )

    def _process_item(self, item: str) -> None:
        """
        Process an input url to be handled by yt-dlp (may be a video or a playlist)
        and pass video urls to the output queue
        """
        logging.debug("Processing url: %s", item)

        # Call yt-dlp in a subprocess
        try:
            completed_process = run(
                (*self._base_command, item),
                capture_output=True,
                check=True,
                encoding="utf-8",
            )
        except CalledProcessError as e:
            logging.error("Failed to get info from url %s: %s", item, e)
            return

        # Extract video URLs (one video infojson per line)
        for output_line in completed_process.stdout.splitlines():
            stripped_line = output_line.strip()
            if len(stripped_line) == 0:
                logging.debug("Empty line from yt-dlp")
                continue
            try:
                video_info_dict = json.loads(stripped_line)
            except json.JSONDecodeError:
                logging.debug("Invalid JSON line from yt-dlp: %s", stripped_line)
                continue
            if not isinstance(video_info_dict, dict):
                logging.debug("Invalid parsed value: %s", video_info_dict)
                continue
            video_url = video_info_dict.get("original_url")
            if video_url is None:
                logging.debug("No video URL in infojson: %s", video_info_dict)
                continue
            logging.debug("Got video URL from yt-dlp: %s", video_url)
            self._send_output(video_url)



import json
import logging
from functools import lru_cache
from multiprocessing import JoinableQueue
from subprocess import PIPE, Popen
from typing import Literal, Sequence, TypedDict

from yt_dlpp.interceptors.interceptor import DlInterceptor
from yt_dlpp.workers.worker import Worker


class _VideoSubdict(TypedDict):
    id: str
    original_url: str
    title: str


class _ProgressSubdict(TypedDict):
    downloaded_bytes: int
    total_bytes: int
    total_bytes_estimate: int
    eta: Literal["NA"] | float
    speed: Literal["NA"] | float
    elapsed: float


class ProgressLineDict(TypedDict):
    video: _VideoSubdict
    progress: _ProgressSubdict


class DownloadWorker(Worker):
    """Worker process that downloads from yt-dlp video urls"""

    input_queue: JoinableQueue
    output_queue: JoinableQueue

    _ydl_args: Sequence[str]

    @property
    @lru_cache(maxsize=1)
    def _base_command(self) -> tuple[str]:
        """Generate the base command"""
        interceptor = DlInterceptor()
        _, allowed = interceptor.parse_known_args(self._ydl_args)
        progress_template = (
            "{"
            + '"video": %(info.{id,original_url,title})j,'
            + '"progress": %(progress.{downloaded_bytes,total_bytes,total_bytes_estimate,eta,speed,elapsed})j'
            + "}"
        )
        return (
            "yt-dlp",
            "--quiet",
            "--progress",
            "--newline",
            "--progress-template",
            progress_template,
            *allowed,
        )

    def __init__(
        self,
        ydl_args: Sequence[str],
        input_queue: JoinableQueue,
        output_queue: JoinableQueue,
    ) -> None:
        super().__init__(input_queue, output_queue)
        self._ydl_args = ydl_args

    def _process_item(self, item: str) -> None:
        # Download the video
        logging.debug("Starting download for %s", item)
        process = Popen(
            (*self._base_command, item),
            encoding="utf-8",
            bufsize=1,
            universal_newlines=True,
            stdout=PIPE,
        )
        # Get progress as soon as a line is available
        for line in process.stdout:
            parsed_line: ProgressLineDict = json.loads(line)
            self._send_output(parsed_line)
        logging.debug("Download finished for %s", item)



import sys
from abc import abstractmethod
from multiprocessing import Process
from typing import Any, Generic, TypeVar

# HACK: Type hints are bad, but it's not my fault.
# mutiprocessing queues don't support type hints, for some god-forsaken reason.
# See https://github.com/python/cpython/issues/99509

TaskInputValueT = TypeVar("TaskInputValueT")
TaskOutputValueT = TypeVar("TaskOutputValueT")


class WorkerInterface(Generic[TaskInputValueT, TaskOutputValueT]):
    """Base class for Worker and WorkerPool"""

    @abstractmethod
    def start(self) -> None:
        """Start the worker"""

    @abstractmethod
    def get_input_queue(self) -> Any:
        """Get the worker's input queue"""

    @abstractmethod
    def dismiss(self) -> None:
        """Signal to the worker to exit"""


class Worker(Process, WorkerInterface[TaskInputValueT, TaskOutputValueT]):
    """Worker process with input and output queues"""

    # --- Protected methods

    def _send_output(self, value: TaskOutputValueT) -> None:
        """Send an item to the output queue if it exists, else do nothing"""
        if self.output_queue is None:
            return
        self.output_queue.put(value)

    @abstractmethod
    def _process_item(self, item: TaskInputValueT) -> None:
        """Process an item and pass results to the output queue"""

    # --- Init

    def __init__(
        self,
        input_queue: Any,
        output_queue: None | Any,
    ) -> None:
        super(Process, self).__init__(daemon=True)
        self.input_queue = input_queue
        self.output_queue = output_queue

    # --- Public methods

    def run(self):
        """Subprocess' main function"""
        while True:
            # Process the next item
            item: TaskInputValueT = self.input_queue.get()
            if item is not None:
                self._process_item(item)
            self.input_queue.task_done()

            # Stop if requested to
            if item is None:
                break

        # Exit gracefuly
        sys.exit(0)

    def start(self):
        super(Process, self).start()

    def dismiss(self):
        self.input_queue.put(None)

    def get_input_queue(self):
        return self.input_queue


class WorkerPool(WorkerInterface[TaskInputValueT, TaskOutputValueT]):
    """Pool of workers sharing an input queue"""

    # --- Protected methods

    __workers: tuple[Worker[TaskInputValueT, TaskOutputValueT]]

    # --- Init

    def __init__(self, *workers: Worker):
        """
        Intitialize a Pool\n
        It is critical that workers share their input queue.
        """
        assert len(workers) > 0, "Cannot create pool with no workers"
        assert (
            len({worker.get_input_queue() for worker in workers}) == 1
        ), "All Workers in a pool must have the same input queue"
        self.__workers = workers

    @classmethod
    def from_class(cls, n: int, klass: type[Worker], *args) -> "WorkerPool":
        """
        Create a worker pool containing n workers of the given class
        with all the same constructor args
        """
        workers = (klass(*args) for _ in range(n))
        return WorkerPool(*workers)

    # --- Public methods

    def start(self):
        for worker in self.__workers:
            worker.start()

    def dismiss(self):
        for worker in self.__workers:
            worker.dismiss()

    def get_input_queue(self):
        return self.__workers[0].get_input_queue()



import logging
from multiprocessing import JoinableQueue
from typing import Optional, TypedDict

from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    filesize,
)

from yt_dlpp.workers.download_worker import ProgressLineDict
from yt_dlpp.workers.worker import Worker


class _VideoTaskInfo(TypedDict):
    """Information about a video task"""

    task_id: TaskID
    started: bool


class _CustomFields(TypedDict):
    """Custom fields for progress tasks"""

    custom_total: str
    custom_speed: str
    custom_eta: str


class ProgressWorker(Worker):
    """Worker in charge of displaying progress info"""

    input_queue: JoinableQueue
    output_queue: None = None

    _progress_bar: Progress
    _tasks: dict[str, _VideoTaskInfo]
    _unknown_value = "?"

    def __init__(self, input_queue: JoinableQueue):
        super().__init__(input_queue, None)

    def run(self) -> None:
        self._tasks = {}
        columns = (
            TextColumn('"[progress.description]{task.description}"'),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("of"),
            TextColumn("{task.fields[custom_total]}", style="progress.filesize"),
            TextColumn("at"),
            TextColumn("{task.fields[custom_speed]}", style="progress.data.speed"),
            TextColumn("ETA"),
            TextColumn("{task.fields[custom_eta]}", style="progress.eta"),
        )
        with Progress(*columns) as self._progress_bar:
            super().run()

    def _get_estimated_total_bytes(
        self, progress_info: ProgressLineDict
    ) -> Optional[float]:
        try:
            return float(progress_info["progress"]["total_bytes_estimate"])
        except (ValueError, KeyError):
            return None

    def _get_real_total_bytes(self, progress_info: ProgressLineDict) -> Optional[float]:
        try:
            return float(progress_info["progress"]["total_bytes"])
        except (ValueError, KeyError):
            return None

    def _get_downloaded_bytes(self, progress_info: ProgressLineDict) -> float:
        try:
            return float(progress_info["progress"]["downloaded_bytes"])
        except (ValueError, KeyError):
            return 0

    def _get_speed(self, progress_info: ProgressLineDict) -> float:
        try:
            return float(progress_info["progress"]["speed"])
        except (ValueError, KeyError):
            return 0

    def _get_eta(self, progress_info: ProgressLineDict) -> Optional[float]:
        try:
            return float(progress_info["progress"]["eta"])
        except (ValueError, KeyError):
            return None

    def _format_custom_total(self, total_bytes: float, is_estimate: bool) -> str:
        # No total bytes available
        if total_bytes is None:
            return self._unknown_value
        # Pretty format total bytes
        prefix = "~ " if is_estimate else ""
        return prefix + filesize.decimal(int(total_bytes))

    def _format_custom_speed(self, speed: float) -> str:
        return filesize.decimal(int(speed)) + "/s"

    def _format_custom_eta(self, eta_seconds: Optional[float]) -> str:
        # No ETA available
        style = "progress.eta"
        if eta_seconds is None:
            return self._unknown_value
        # Pretty format ETA
        minutes, seconds = divmod(int(eta_seconds), 60)
        hours, minutes = divmod(minutes, 60)
        return (
            f"{hours:d}:{minutes:02d}:{seconds:02d}"
            if hours
            else f"{minutes:02d}:{seconds:02d}"
        )

    def _create_empty_custom_fields(self) -> _CustomFields:
        return {
            "custom_total": self._format_custom_total(None, True),
            "custom_speed": self._format_custom_speed(0),
            "custom_eta": self._format_custom_eta(None),
        }

    def _task_exists(self, video_id: str) -> bool:
        """Check if a progress task exists for a video ID"""
        return video_id in self._tasks

    def _create_task(self, video_id: str, title: str) -> None:
        """Create a new progress task for a video ID"""
        logging.debug("Creating progress task for %s", video_id)
        task_id = self._progress_bar.add_task(
            description=title, start=False, **self._create_empty_custom_fields()
        )
        self._tasks[video_id] = _VideoTaskInfo(task_id=task_id, started=False)

    def _task_started(self, video_id: str) -> bool:
        """Check if the progress task has been started for a video ID"""
        return self._tasks[video_id]["started"]

    def _start_task(self, video_id: str) -> None:
        """Start the progress task for a video ID"""
        logging.debug("Starting progress task for %s", video_id)
        task_id = self._tasks[video_id]["task_id"]
        self._progress_bar.start_task(task_id)
        self._tasks[video_id]["started"] = True

    def _update_task(
        self,
        video_id: str,
        downloaded_bytes: float,
        total_bytes: float,
        total_is_estimate: bool,
        speed_bytes_per_second: float,
        eta_seconds: float,
    ) -> None:
        """Update the progress task for a video ID"""
        task_id = self._tasks[video_id]["task_id"]
        fields = {
            "custom_total": self._format_custom_total(total_bytes, total_is_estimate),
            "custom_speed": self._format_custom_speed(speed_bytes_per_second),
            "custom_eta": self._format_custom_eta(eta_seconds),
        }
        self._progress_bar.update(
            task_id, completed=downloaded_bytes, total=total_bytes, **fields
        )

    def _process_item(self, progress_info: ProgressLineDict) -> None:
        # Get current info
        video_id = progress_info["video"]["id"]
        real_total_bytes = self._get_real_total_bytes(progress_info)
        estimated_total_bytes = self._get_estimated_total_bytes(progress_info)
        total_bytes = real_total_bytes or estimated_total_bytes
        downloaded_bytes = self._get_downloaded_bytes(progress_info)
        speed_bytes_per_second = self._get_speed(progress_info)
        eta_seconds = self._get_eta(progress_info)
        # Create / start / update task
        if not self._task_exists(video_id):
            self._create_task(video_id, progress_info["video"]["title"])
        if not self._task_started(video_id) and total_bytes is not None:
            self._start_task(video_id)
        if self._task_started(video_id):
            self._update_task(
                video_id,
                downloaded_bytes,
                total_bytes,
                real_total_bytes is None,
                speed_bytes_per_second,
                eta_seconds,
            )



from argparse import ArgumentParser, Namespace
from typing import Optional, Sequence


class AbstractInterceptor(ArgumentParser):
    """Abstract class for intercepting arguments"""

    def __init__(self) -> None:
        super().__init__(add_help=False, allow_abbrev=False)

    def add_intercepted_arguments(self, *args: str | tuple[str]):
        for arg_or_argtuple in args:
            match arg_or_argtuple:
                case str():
                    self.add_argument(arg_or_argtuple)
                case tuple():
                    self.add_argument(*arg_or_argtuple)


class InputUrlInterceptorNamespace(Namespace):
    """Namespace for yt-dlp URLs interceptor"""

    batch_file: str | None
    urls: list[str]


class InputUrlsInterceptor(AbstractInterceptor):
    """Parser to intercept yt-dlp URLs"""

    def __init__(self) -> None:
        super().__init__()
        self.add_intercepted_arguments("--batch-file")
        self.add_argument("urls", nargs="*")

    def parse_known_args(
        self,
        args: Optional[Sequence[str]] = None,
        namespace: Optional[Namespace] = None,
    ) -> tuple[InputUrlInterceptorNamespace, list[str]]:
        return super().parse_known_args(args, namespace)


class _AppInterceptor(AbstractInterceptor):
    """Parser to intercept aruments that are not allowed throughout the app"""

    def __init__(self) -> None:
        super().__init__()
        self.add_intercepted_arguments(
            ("--help", "-h"),
            "--version",
            ("--update", "-U"),
            "--update-to",
            "--dump-user-agent",
            "--list-extractors",
            "--extractor-descriptions",
            "--alias",
            "--batch-file",  # This is intercepted by InputUrlInterceptor only
        )


class InfoInterceptor(_AppInterceptor):
    """Parser to intercept disallowed arguments for the info worker"""

    def __init__(self) -> None:
        super().__init__()
        self.add_intercepted_arguments(
            "--no-simulate",
            "--dump-single-json",
        )


class DlInterceptor(_AppInterceptor):
    """Parser to intercept disallowed arguments for the download worker"""

    def __init__(self) -> None:
        super().__init__()
        self.add_intercepted_arguments(
            "--no-progress",
            "--console-title",
            "--progress-template",
            ("--quiet", "-q"),
            ("--print", "-O"),
            ("--verbose", "-v"),
            "--dump-pages",
            "--print-traffic",
        )



import logging
import sys
from argparse import ArgumentParser, Namespace
from multiprocessing import JoinableQueue, cpu_count
from os import getenv
from typing import Optional, Sequence

from yt_dlpp.interceptors.interceptor import InputUrlsInterceptor
from yt_dlpp.workers.dedup_worker import DedupWorker
from yt_dlpp.workers.download_worker import DownloadWorker
from yt_dlpp.workers.info_worker import InfoWorker
from yt_dlpp.workers.progress_worker import ProgressWorker
from yt_dlpp.workers.worker import WorkerInterface, WorkerPool


def _setup_logging() -> None:
    """Setup the logging"""
    log_levels = logging.getLevelNamesMapping()
    log_level = log_levels[getenv("LOG_LEVEL", "ERROR")]
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - [%(processName)s - %(levelname)s] %(message)s",
    )


class YtdlppParserNamespace(Namespace):
    """Namespace for yt-dlpp parser args"""

    n_info_workers: int
    n_dl_workers: int


class YtdlppParser(ArgumentParser):
    """Parser for yt-dlpp"""

    def __init__(self) -> None:
        super().__init__(
            description="A wrapper to download content using yt-dlp in parallel",
            epilog="See `yt-dlp --help` for more CLI options",
            allow_abbrev=False,
        )
        self.add_argument(
            "--n-info-workers",
            type=int,
            default=cpu_count(),
            help="Number of info workers to use",
        )
        self.add_argument(
            "--n-dl-workers",
            type=int,
            default=cpu_count(),
            help="Number of download workers to use",
        )

    def parse_known_args(
        self,
        args: Optional[Sequence[str]] = None,
        namespace: Optional[Namespace] = None,
    ) -> tuple[YtdlppParserNamespace, list[str]]:
        return super().parse_known_args(args, namespace)


def main():
    """App entry point"""

    # Enable logging to be able to debug if needed
    _setup_logging()

    # Parse the main arguments
    logging.debug("Parsing yt-dlpp args")
    args, raw_ydtdlp_args = YtdlppParser().parse_known_args()

    # Intercept some yt-dlp args
    logging.debug("Intercepting yt-dlp arguments")
    input_urls_args, ytdlp_args = InputUrlsInterceptor().parse_known_args(
        raw_ydtdlp_args
    )

    # Get the input URLs
    urls = []
    if input_urls_args.urls:
        urls.extend(input_urls_args.urls)
    if input_urls_args.batch_file:
        logging.debug("Reading URLs from batch file: %s", input_urls_args.batch_file)
        try:
            with open(input_urls_args.batch_file, "r") as file:
                lines = file.readlines()
            urls.extend({url for line in lines if (url := line.strip())})
        except OSError as e:
            logging.error("Error reading batch file: %s", e)
    if len(urls) == 0:
        logging.error("No URLs to process")
        sys.exit(1)

    # Create the queues
    logging.debug("Creating queues")
    input_url_queue = JoinableQueue()
    video_url_queue = JoinableQueue()
    unique_video_url_queue = JoinableQueue()
    progress_queue = JoinableQueue()

    # Create the workers
    logging.debug("Creating workers")
    workers: tuple[WorkerInterface] = (
        WorkerPool.from_class(
            args.n_info_workers,
            InfoWorker,
            ytdlp_args,
            input_url_queue,
            video_url_queue,
        ),
        DedupWorker(
            video_url_queue,
            unique_video_url_queue,
        ),
        WorkerPool.from_class(
            args.n_dl_workers,
            DownloadWorker,
            ytdlp_args,
            unique_video_url_queue,
            progress_queue,
        ),
        ProgressWorker(
            progress_queue,
        ),
    )

    # Start the workers
    logging.debug("Starting workers")
    for worker in workers:
        worker.start()

    # Send the initial URLs to the queue
    print("Getting video info...")
    logging.debug("Sending initial URLs to the queue")
    for url in urls:
        logging.debug("\t %s", url)
        input_url_queue.put(url)

    # Wait for every step to finish, one after the other
    for i, worker in enumerate(workers):
        kind = "WorkerPool" if isinstance(worker, WorkerPool) else "Worker"
        logging.debug("Waiting for %s %d to finish", kind, i)
        worker.dismiss()
        logging.debug("Dismissed %s %d", kind, i)
        worker_input_queue = worker.get_input_queue()
        worker_input_queue.close()
        worker_input_queue.join()
        logging.debug("%s %d finished", kind, i)
    logging.debug("All workers finished")

    # If all went well, all of our workers finished
    # The remaining ones will be killed at exit since they're daemon processes
    sys.exit(0)


if __name__ == "__main__":
    main()
